﻿'
' Copyright (c) 2021, MindFusion LLC - Bulgaria.
'


''' <summary>
''' Interaction logic for MainWindow.xaml
''' </summary>
Partial Public Class MainWindow
	Inherits Window
	Public Sub New()
		InitializeComponent()

		diagram.Selection.AllowMultipleSelection = False
		diagram.BackBrush = New LinearGradientBrush(Colors.White, Colors.DodgerBlue, 90)
		Dim glass As New GlassEffect()
		glass.Type = GlassEffectType.Type2
		diagram.NodeEffects.Add(glass)

		' add object and process shapes to the NodeListView control
		AddToPalette(objectShape, "Object")
		AddToPalette(processShape, "Process")

		' create two default objects
		For i As Integer = 0 To 1
			Dim node = New ShapeNode(diagram) With { _
			 .HandlesStyle = HandlesStyle.DashFrame, _
			 .Text = "Object " & (i + 1), _
			 .AllowOutgoingLinks = False, _
			 .AllowIncomingLinks = False, _
			 .Bounds = New Rect((44 * mm) * i + 10, 10, 24 * mm, 24 * mm), _
			 .Shape = objectShape, _
			 .EnabledHandles = AdjustmentHandles.None, _
			 .Brush = New LinearGradientBrush(Colors.White, Colors.LimeGreen, 90) _
			}
			objects.Add(node)
			diagram.Items.Add(node)
		Next

		' add sequence diagram lifelines extending below the object nodes
		AddLifelines()
	End Sub

	Private Sub AddToPalette(ByVal shape As Shape, ByVal label As String)
		Dim n1 = New ShapeNode(diagram) With { _
		 .Shape = shape, _
		 .Bounds = New Rect(0, 0, 25, 25), _
		 .Brush = New SolidColorBrush(Color.FromArgb(255, 0, 255, 255)) _
		}
		nodeListView.Items.Add(n1)
		NodeListView.SetLabel(n1, label)
	End Sub

	Private Sub OnNodeCreated(ByVal sender As Object, ByVal e As NodeEventArgs)
		Dim node = DirectCast(e.Node, ShapeNode)
		node.HandlesStyle = HandlesStyle.DashFrame

		If node.Shape Is objectShape Then
			node.AllowOutgoingLinks = False
			node.AllowIncomingLinks = False
			node.EnabledHandles = AdjustmentHandles.None
			node.Brush = New LinearGradientBrush(Colors.White, Colors.LimeGreen, 90)
			objects.Add(node)
			node.Text = "Object " & System.Convert.ToString(objects.Count)

			AddLifelines()
		End If

		If node.Shape Is processShape Then
			node.Brush = New LinearGradientBrush(Colors.White, Colors.Red, 90)
			node.TextRotationAngle = -90
			node.PolygonalTextLayout = True
			node.TextAlignment = TextAlignment.Center
			node.TextVerticalAlignment = AlignmentY.Center
			processes.Add(node)
			node.Text = "Process " & System.Convert.ToString(processes.Count)
			node.EnabledHandles = AdjustmentHandles.ResizeBottomCenter Or AdjustmentHandles.Move Or AdjustmentHandles.ResizeTopCenter
			node.Constraints.MoveDirection = DirectionConstraint.Vertical

			' use grouping to associate this process node with its lifeline
			Dim lifeline As DiagramLink = NearestLifeline(node)
			node.AttachTo(lifeline, AttachToLink.Point, 0)
			node.Tag = lifeline.Origin
			node.Bounds = New Rect(lifeline.StartPoint.X - 6 * mm, node.Bounds.Y, 12 * mm, 24 * mm)
		End If

		diagram.EndEdit(True)
		diagram.BeginEdit(node)

		' grow the diagram to the right if necessary
		Dim contentRect = diagram.GetContentBounds(False, False)
		If contentRect.Right > diagram.Bounds.Right Then
			Dim r = diagram.Bounds
			r.Width = contentRect.Right - r.Left + 30
			diagram.Bounds = r
		End If
	End Sub

	Private Sub OnNodeModified(ByVal sender As Object, ByVal e As NodeEventArgs)
		' after this node is modified, check if it overlaps the next node
		' on the lifeline and push nodes down if necessary
		If e.Node.MasterGroup IsNot Nothing AndAlso TypeOf e.Node.MasterGroup.MainItem Is DiagramLink Then
			Dim group = e.Node.MasterGroup
			Dim nodes = New List(Of DiagramNode)(group.AttachedNodes.Cast(Of DiagramNode)())
			nodes.Sort(Function(n1, n2) n1.Bounds.Top.CompareTo(n2.Bounds.Top))
			Dim nodesBelow = nodes.Skip(nodes.IndexOf(e.Node) + 1)
			If nodesBelow.Count() > 0 AndAlso nodesBelow.First().Bounds.Top <= e.Node.Bounds.Bottom Then
				Dim offset = e.Node.Bounds.Bottom - nodesBelow.First().Bounds.Top + 30
				For Each node As DiagramNode In nodesBelow
					node.Move(node.Bounds.X, node.Bounds.Y + offset)
				Next
			End If
		End If

		ArrangeLinks()
	End Sub

	Private Sub OnLinkCreated(ByVal sender As Object, ByVal e As LinkEventArgs)
		ArrangeLinks()
	End Sub

	' distributes message links between processes so that they are horizontal and
	' do not overlap when there are more than one links connecting the same pair
	Private Sub ArrangeLinks()
		Dim bundles = FindLinkBundles()
		For Each bundle As DiagramLinkCollection In bundles.Values
			Dim range = SharedRange(bundle(0).Origin, bundle(0).Destination)
			If range IsNot Nothing Then
				Dim h As Double = range(1) - range(0)
				Dim s As Double = h / (bundle.Count + 1)
				Dim y As Double = range(0) + s
				For Each link As DiagramLink In bundle
					link.Shape = LinkShape.Polyline
					link.SegmentCount = 1
					link.ReassignAnchorPoints()

					link.StartPoint = New Point(link.StartPoint.X, y)
					link.EndPoint = New Point(link.EndPoint.X, y)
					link.UpdateFromPoints()

					y += s
				Next
			Else
				diagram.RoutingOptions.StartOrientation = Orientation.Horizontal
				diagram.RoutingOptions.EndOrientation = Orientation.Horizontal

				Dim l As DiagramLink = bundle(0)
				Dim left As DiagramNode, right As DiagramNode
				If l.Origin.Bounds.X < l.Destination.Bounds.X Then
					left = l.Origin
					right = l.Destination
				Else
					left = l.Destination
					right = l.Origin
				End If

				Dim ly As Double, ry As Double, h As Double = 20, s As Double = h / (bundle.Count + 1)
				If left.Bounds.Bottom < right.Bounds.Top Then
					ly = s + left.Bounds.Bottom - h
					ry = s + right.Bounds.Top
				Else
					ly = s + left.Bounds.Top
					ry = s + right.Bounds.Bottom - h
				End If
				For Each link As DiagramLink In bundle
					link.Shape = LinkShape.Cascading
					If link.Origin Is left Then
						link.StartPoint = New Point(left.Bounds.Right, ly)
						link.EndPoint = New Point(right.Bounds.Left, ry)
					Else
						link.StartPoint = New Point(right.Bounds.Left, ry)
						link.EndPoint = New Point(left.Bounds.Right, ly)
					End If
					ly += s
					ry += s
				Next
				diagram.LinkRouter.RouteLinks(bundle)
			End If
		Next
	End Sub

	Private Function NearestLifeline(ByVal node As DiagramNode) As DiagramLink
		Dim nodeX As Double = node.GetCenter().X
		Dim minDist As Double = Double.MaxValue
		Dim nearest As DiagramLink = Nothing

		For Each link As DiagramLink In lifelines
			Dim distance As Double = Math.Abs(link.StartPoint.X - nodeX)
			If distance < minDist Then
				minDist = distance
				nearest = link
			End If
		Next

		Return nearest
	End Function

	Private Sub AddLifelines()
		For i As Integer = 0 To objects.Count - 1
			If objects(i).OutgoingLinks.Count = 0 Then
				Dim lifeline = New DiagramLink(diagram) With { _
				 .Tag = "Lifeline", _
				 .Origin = objects(i), _
				 .Destination = diagram.Dummy, _
				 .Locked = True, _
				 .Stroke = New SolidColorBrush(Color.FromArgb(255, 50, 50, 50)), _
				 .StrokeThickness = 2, _
				 .StrokeDashStyle = DashStyles.Dash, _
				 .HeadShape = Nothing _
				}
				diagram.Links.Add(lifeline)
				lifelines.Add(lifeline)
			End If
		Next

		CompactLifelines()
	End Sub

	Private Sub CompactLifelines()
		For i As Integer = 0 To objects.Count - 1
			objects(i).Bounds = New Rect((44 * mm) * i + 10, 10, 24 * mm, 24 * mm)

			Dim objectX As Double = objects(i).GetCenter().X
			Dim lifeline = objects(i).OutgoingLinks(0)
			Dim startPoint = New Point(objectX, objects(i).Bounds.Bottom)
			Dim endPoint = New Point(objectX, startPoint.Y + diagram.Bounds.Height)
			If startPoint <> lifeline.StartPoint OrElse endPoint <> lifeline.EndPoint Then
				lifeline.StartPoint = startPoint
				lifeline.EndPoint = endPoint
				lifeline.UpdateFromPoints()

				If lifeline.SubordinateGroup IsNot Nothing Then
					For Each process As DiagramNode In lifeline.SubordinateGroup.AttachedNodes
						process.Move(objectX - process.Bounds.Width / 2, process.Bounds.Y)
					Next
				End If
			End If
		Next
	End Sub

	Private Sub OnNodeDoubleClicked(ByVal sender As Object, ByVal e As NodeEventArgs)
		diagram.EndEdit(True)
		diagram.BeginEdit(TryCast(e.Node, ShapeNode))
	End Sub

	Private Sub OnNodeDeleted(ByVal sender As Object, ByVal e As NodeEventArgs)
		For i As Integer = diagram.Items.Count - 1 To 0 Step -1
			If diagram.Items(i).Tag Is e.Node Then
				diagram.Items.Remove(diagram.Items(i))
			End If
		Next

		If objects.Contains(e.Node) Then
			objects.Remove(e.Node)
		End If
		If processes.Contains(e.Node) Then
			processes.Remove(e.Node)
		End If

		Dim node = DirectCast(e.Node, ShapeNode)
		If node.Shape Is objectShape Then
			CompactLifelines()
		End If
	End Sub

	Private Sub OnEnterInplaceEditMode(ByVal sender As Object, ByVal e As InPlaceEditEventArgs)
		e.TextBox.TextAlignment = TextAlignment.Center
		e.TextBox.VerticalContentAlignment = VerticalAlignment.Center

		e.TextBox.SelectAll()

		Dim node = TryCast(e.Node, ShapeNode)
		If node Is Nothing OrElse node.Shape IsNot processShape Then
			Return
		End If

		' display the edit box of process nodes rotated to match
		' how their text will appear with TextRotationAngle set
		Dim nodeWidth As Double = e.Node.Bounds.Width
		Dim nodeHeight As Double = e.Node.Bounds.Height
		e.TextBox.Width = nodeHeight
		e.TextBox.Height = nodeWidth

		Dim rotation = New Matrix()
		rotation.Translate(nodeWidth / 2 - nodeHeight / 2, nodeHeight / 2 - nodeWidth / 2)
		rotation.RotateAt(-90, nodeWidth / 2, nodeHeight / 2)
		rotation.Append(e.TextBox.RenderTransform.Value)
		e.TextBox.RenderTransform = New MatrixTransform(rotation)
	End Sub

	Private Shadows Sub OnDragOver(ByVal sender As Object, ByVal e As DragEventArgs)
		If nodeListView.SelectedIndex = 0 Then
			Return
		End If

		e.Handled = True

		' allow dropping process nodes only if they are near a lifeline
		Dim mousePos As Point = e.GetPosition(diagram.DocumentPlane)
		For Each node As DiagramNode In objects
			If node.Bounds.Left < mousePos.X AndAlso node.Bounds.Right > mousePos.X Then
				e.Effects = DragDropEffects.Copy
				Return
			End If
		Next

		e.Effects = DragDropEffects.None
	End Sub

	Private Sub OnNodeModifying(ByVal sender As Object, ByVal e As NodeValidationEventArgs)
		e.Cancel = e.Node.Bounds.Top < diagram.Nodes(0).Bounds.Bottom
	End Sub

	Private Sub OnLinkDeleted(ByVal sender As Object, ByVal e As LinkEventArgs)
		If lifelines.Contains(e.Link) Then
			lifelines.Remove(e.Link)
		End If
	End Sub

	Private Sub OnLinkCreating(ByVal sender As Object, ByVal e As LinkValidationEventArgs)
		' do not allow links between processes on the same lifeline
		If e.Destination IsNot Nothing Then
			Dim idx1 As Integer = lifelines.IndexOf(TryCast(e.Origin.MasterGroup.MainItem, DiagramLink))
			Dim idx2 As Integer = lifelines.IndexOf(TryCast(e.Destination.MasterGroup.MainItem, DiagramLink))
			e.Cancel = idx1 = idx2
		End If
	End Sub

	Private Function PairCode(ByVal link As DiagramLink) As Integer
		Dim i1 As Integer = link.Origin.ZIndex
		Dim i2 As Integer = link.Destination.ZIndex
		Return diagram.Items.Count * Math.Max(i1, i2) + Math.Min(i1, i2)
	End Function

	Private Function FindLinkBundles() As Dictionary(Of Integer, DiagramLinkCollection)
		Dim bundles = New Dictionary(Of Integer, DiagramLinkCollection)()
		For Each link As DiagramLink In diagram.Links
			If "Lifeline".Equals(link.Tag) Then
				Continue For
			End If

			Dim code As Integer = PairCode(link)
			If Not bundles.ContainsKey(code) Then
				bundles(code) = New DiagramLinkCollection()
			End If
			bundles(code).Add(link)
		Next
		Return bundles
	End Function

	Private Function SharedRange(ByVal n1 As DiagramNode, ByVal n2 As DiagramNode) As Double()
		Dim r1 = n1.Bounds
		Dim r2 = n2.Bounds

		If r1.Top > r2.Bottom Then
			Return Nothing
		End If
		If r1.Bottom < r2.Top Then
			Return Nothing
		End If

		Return New Double() {Math.Max(r1.Top, r2.Top), Math.Min(r1.Bottom, r2.Bottom)}
	End Function

	ReadOnly objectShape As Shape = Shapes.MultiProc
	ReadOnly processShape As Shape = Shapes.Procedure

	ReadOnly objects As New List(Of DiagramNode)()
	ReadOnly processes As New List(Of DiagramNode)()
	ReadOnly lifelines As New List(Of DiagramLink)()

	Const mm As Double = 4
End Class